package org.jbox2d.common;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import org.jbox2d.collision.shapes.CircleShape;
import org.jbox2d.collision.shapes.PolygonShape;
import org.jbox2d.dynamics.Body;
import org.jbox2d.dynamics.FixtureDef;
import org.json.JSONArray;
import org.json.JSONException;
import org.json.JSONObject;
import org.json.JSONTokener;

import anywheresoftware.b4a.BA;
import anywheresoftware.b4a.objects.collections.List;
import anywheresoftware.b4a.objects.streams.File;
import anywheresoftware.b4a.objects.streams.File.InputStreamWrapper;
import anywheresoftware.b4a.objects.streams.File.TextReaderWrapper;

//Author: Aurélien Ribbon & F. Leneuf-Magaud
@BA.ShortName("jbBodyEditorLoader")
public class BodyEditorLoaderB4A
{
	// Model
	private Model model = null;

	// Reusable stuff
	private final ArrayList<Vec2> vectorPool = new ArrayList<Vec2>();
	private final PolygonShape polygonShape = new PolygonShape();
	private final CircleShape circleShape = new CircleShape();
	private final Vec2 vec = new Vec2();

	// -------------------------------------------------------------------------
	// Ctors
	// -------------------------------------------------------------------------

	public BodyEditorLoaderB4A() {}

	public void Load(String Dir, String FileName) throws IOException, JSONException
	{
		if (Dir.equals("") || FileName.equals(""))
			throw new NullPointerException("Dir or FileName is empty");
		final InputStreamWrapper fileIS = File.OpenInput(Dir, FileName);
		final TextReaderWrapper fileJSON = new TextReaderWrapper();
		fileJSON.Initialize(fileIS.getObject());
		final String str = fileJSON.ReadAll();
		fileJSON.Close();
		model = readJson(str);
	}

	// -------------------------------------------------------------------------
	// Public API
	// -------------------------------------------------------------------------

	/**
	 * Creates and applies the fixtures defined in the editor. The name
	 * parameter is used to retrieve the right fixture from the loaded file.
	 *
	 * The body reference point (the red cross in the tool) is by default
	 * located at the bottom left corner of the image. This reference point
	 * will be put right over the BodyDef position point. Therefore, you should
	 * place this reference point carefully to let you place your body in your
	 * world easily with its BodyDef.position point. Note that to draw an image
	 * at the position of your body, you will need to know this reference point
	 * (see GetOrigin).
	 *
	 * Also, saved shapes are normalized. As shown in the tool, the width of
	 * the image is considered to be always 1 meter. Thus, you need to provide
	 * a scale factor so the polygons get resized according to your needs (not
	 * every body is 1 meter large in your game, I guess).
	 *
	 * @param body The JBox2D body you want to attach the fixture to.
	 * @param name The name of the fixture you want to load.
	 * @param fd The fixture parameters to apply to the created body fixture.
	 * @param scale The desired scale of the body. The default width is 1.
	 */
	public void AttachFixture(Body body, String name, FixtureDef fd, float scale) {
		RigidBodyModel rbModel = model.rigidBodies.get(name);
		if (rbModel == null)
			throw new RuntimeException("Name '" + name + "' was not found.");

		Vec2 origin = vec.set(rbModel.origin).mul(scale);

		for (int i=0, n=rbModel.polygons.size(); i<n; i++) {
			PolygonModel polygon = rbModel.polygons.get(i);
			Vec2[] vertices = polygon.buffer;

			for (int ii=0, nn=vertices.length; ii<nn; ii++) {
				vertices[ii] = newVec().set(polygon.vertices.get(ii)).mul(scale);
				vertices[ii].sub(origin);
			}

			polygonShape.set(vertices, vertices.length);
			fd.shape = polygonShape;
			body.createFixture(fd);

			for (int ii=0, nn=vertices.length; ii < nn; ii++) {
				free(vertices[ii]);
			}
		}

		for (int i=0, n=rbModel.circles.size(); i < n; i++) {
			CircleModel circle = rbModel.circles.get(i);
			Vec2 center = newVec().set(circle.center).mul(scale);
			float radius = circle.radius * scale;

			circleShape.setCenterPosition(center.x, center.y);
			circleShape.setRadius(radius);
			fd.shape = circleShape;
			body.createFixture(fd);

			free(center);
		}
	}

	/**
	 * Gets the image path attached to the given name.
	 */
	public String GetImagePath(String name) {
		RigidBodyModel rbModel = model.rigidBodies.get(name);
		if (rbModel == null)
			throw new RuntimeException("Name '" + name + "' was not found.");

		return rbModel.imagePath;
	}

	/**
	 * Gets the origin point attached to the given name. Since the point is
	 * normalized in [0,1] coordinates, it needs to be scaled to your body
	 * size. Warning: this method returns the same Vec2 object each time, so
	 * copy it if you need it for later use.
	 */
	public Vec2 GetOrigin(String name, float scale) {
		RigidBodyModel rbModel = model.rigidBodies.get(name);
		if (rbModel == null)
			throw new RuntimeException("Name '" + name + "' was not found.");

		return vec.set(rbModel.origin).mul(scale);
	}

	/**
	 * Gets a sorted list of all names.
	 */
	public List SortedList()
	{
		List L = new List();
		L.Initialize();
		for (String key : model.rigidBodies.keySet()) {
			L.Add(key);
		}
		L.Sort(true);
		return L;
	}

	/**
	 * <b>For advanced users only.</b> Lets you access the internal model of
	 * this loader and modify it. Be aware that any modification is permanent
	 * and that you should really know what you are doing.
	 */
	//public Model getInternalModel() {
	//	return model;
	//}

	// -------------------------------------------------------------------------
	// Json Models
	// -------------------------------------------------------------------------

	public static class Model {
		public final HashMap<String, RigidBodyModel> rigidBodies = new HashMap<String, RigidBodyModel>();
	}

	public static class RigidBodyModel {
		public String name;
		public String imagePath;
		public final Vec2 origin = new Vec2();
		public final ArrayList<PolygonModel> polygons = new ArrayList<PolygonModel>();
		public final ArrayList<CircleModel> circles = new ArrayList<CircleModel>();
	}

	public static class PolygonModel {
		public final ArrayList<Vec2> vertices = new ArrayList<Vec2>();
		private Vec2[] buffer; // used to avoid allocation in attachFixture()
	}

	public static class CircleModel {
		public final Vec2 center = new Vec2();
		public float radius;
	}

	// -------------------------------------------------------------------------
	// Json reading process
	// -------------------------------------------------------------------------

	private LinkedHashMap<String, Object> parseJsonObject(Object o) throws JSONException
	{
		LinkedHashMap<String, Object> map = new LinkedHashMap<String, Object>(); 
		JSONObject jo = (JSONObject)o;
		Iterator<?> it = jo.keys();
		while (it.hasNext()) {
			String key = (String)it.next();
			Object value = jo.get(key);
			map.put(key, value);
		}
		return map;
	}

	private LinkedHashMap<String, Object> parseJsonArray(Object o) throws JSONException
	{
		LinkedHashMap<String, Object> map = new LinkedHashMap<String, Object>(); 
		JSONArray ja = (JSONArray)o;
		for(int i = 0; i < ja.length(); i++) {
			map.put(String.valueOf(i), ja.get(i));
		}
		return map;
	}

	private Model readJson(String str) throws JSONException {
		Model m = new Model();

		JSONTokener j = new JSONTokener(str);
		Object o = j.nextValue();
		if (!(o instanceof JSONObject)) {
			throw new RuntimeException("JSON Object expected.");
		}

		LinkedHashMap<String, Object> bodiesElems = parseJsonArray(parseJsonObject(o).get("rigidBodies"));
		for (String key : bodiesElems.keySet()) {
			RigidBodyModel rbModel = readRigidBody(bodiesElems.get(key));
			m.rigidBodies.put(rbModel.name, rbModel);
		}

		return m;
	}

	private float toFloat(Object o)
	{
		if (o instanceof Integer) {
			int i = (Integer) o;
			return (float) i;
		}
		else if (o instanceof Double) {
			double d = (Double) o;
			return (float) d;
		}
		return (Float) o;
	}

	private RigidBodyModel readRigidBody(Object bodyElemObj) throws JSONException {
		LinkedHashMap<String, Object> bodyElem = parseJsonObject(bodyElemObj);
		RigidBodyModel rbModel = new RigidBodyModel();
		rbModel.name = (String) bodyElem.get("name");
		rbModel.imagePath = (String) bodyElem.get("imagePath");

		LinkedHashMap<String, Object> originElem = parseJsonObject(bodyElem.get("origin"));
		rbModel.origin.x = toFloat(originElem.get("x"));
		rbModel.origin.y = toFloat(originElem.get("y"));

		// polygons

		LinkedHashMap<String, Object> polygonsElem = parseJsonArray(bodyElem.get("polygons"));
		for (String keyP : polygonsElem.keySet()) {
			PolygonModel polygon = new PolygonModel();
			rbModel.polygons.add(polygon);

			LinkedHashMap<String, Object> verticesElem = parseJsonArray(polygonsElem.get(keyP));
			for (String keyV : verticesElem.keySet()) {
				LinkedHashMap<String,?> vertexElem = parseJsonObject(verticesElem.get(keyV));
				polygon.vertices.add(new Vec2(toFloat(vertexElem.get("x")), toFloat(vertexElem.get("y"))));
			}

			polygon.buffer = new Vec2[polygon.vertices.size()];
		}

		// circles

		LinkedHashMap<String, Object> circlesElem = parseJsonArray(bodyElem.get("circles"));
		for (String keyC : circlesElem.keySet()) {
			CircleModel circle = new CircleModel();
			rbModel.circles.add(circle);

			LinkedHashMap<String,?> circleElem = parseJsonObject(circlesElem.get(keyC));
			circle.center.x = toFloat(circleElem.get("cx"));
			circle.center.y = toFloat(circleElem.get("cy"));
			circle.radius = toFloat(circleElem.get("r"));
		}

		return rbModel;
	}

	// -------------------------------------------------------------------------
	// Helpers
	// -------------------------------------------------------------------------

	private Vec2 newVec() {
		return vectorPool.isEmpty() ? new Vec2() : vectorPool.remove(0);
	}

	private void free(Vec2 v) {
		vectorPool.add(v);
	}
}